# Copyright 2025 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from lightworks import State


def convert_fock_data_to_state(raw_state: str) -> State:
    """
    Takes raw fock state data and converts this to a Lightworks state.
    """
    try:
        if raw_state[0] == "(":
            raw_state = raw_state[1:-1]
        mode_data = raw_state.split(", ")
        if len(mode_data) == 1:
            mode_data[0] = mode_data[0].replace(",", "")
        state = State([int(s) for s in mode_data if s])
    except Exception as e:
        raise ValueError(
            "Unable to convert the provided State to a Lightworks state."
        ) from e
    return state
